#ifndef _LIBDEMOSAICKING_H_
#define _LIBDEMOSAICKING_H_


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "libraryBasic.h"




namespace libIIPStable 
{
	
    
    
    
    

    
    
    
#define VERTICAL 1
#define HORIZONTAL 0
    
    
    
    /**
     * @file   libdemosaicking.h
     * @brief  Demosaicking library
     *
     * 
     *
     * @author Antoni Buades <toni.buades@uib.es>
     */
    
    
    void demosaicking_init_cfa(int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height);
    
    
    
    /**
     * \brief  Classical Adams-Hamilton demosaicking algorithm
     *
     *  The green channel is interpolated directionally depending on the green first and red and blue second directional derivatives.
     *  The red and blue differences with the green channel are interpolated bilinearly.
     *
     */
    
    
    void demosaicking_adams(float threshold, int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height);
    
    
    
    /**
     * \brief  Classical Adams-Hamilton demosaicking algorithm
     *
     *  The green channel is interpolated directionally depending on the value of the parameter direction
     *  The red and blue differences with the green channel are interpolated bilinearly.
     *
     */
    
    
    void demosaicking_adams_directional(int direction, int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height);
    
    
    void demosaicking_adams_chromatic_regularity(int side, int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height);
    
    
    
    /**
     * \brief  Classical bilinear interpolation of red and blue differences with the green channel
     *
     */
    
    void demosaicking_bilinear_red_blue(int redx,int redy,float *ored,float *ogreen,float *oblue,int width,int height);
    
    
    
    
    /**
     * \brief  NLmeans based demosaicking
     */
    
    
    void demosaicking_nlmeans(int bloc, float h,int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height);
    
    
    
    
    /**
     * \brief  NLmeans based demosaicking for a single frame of a sequence
     */
    
    void demosaicking_nlmeans_single_frame(int iIndex,	// iIndex of denoised frame
                                           int redx, int redy,	// red pixel position
                                           int iDtemp,	// Half size of temporal window
                                           int iDWin,		// Half size of comparison window  
                                           int iDBloc,		// Half size of research window
                                           float fFiltPar,	// Filtering parameter
                                           float ***fpI,		// Input
                                           float **fpO,		// Output
                                           int iFrames, int iChannels, int iWidth,int iHeight);
    
    
    /**
     * \brief  Iterate median filter on chromatic components of the image
     *
     */
    
    
    
    void chromatic_median(int iter,int redx,int redy,int projflag,float side,float *ired,float *igreen, float *iblue,float *ored,float *ogreen,float *oblue,int width,int height);
    
    
    
    
    /**
     * \brief Demosaicking chain
     *
     */
    
    
    void ssd_demosaicking_chain(int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height);
    
    
    
}


#endif
