#include "libdenoising.h"



namespace libIIPStable 
{
	
	
	
		
	
    
    
    
	double scalar_product(double *u, double *v, int n)
	{
		
		double aux = 0.0f;
		for(int i=0; i < n; i++)
			aux += u[i] * v[i];
		
		return aux;
		
	}
	
	
	float scalar_product(float *u, float *v, int n)
	{
		
		float aux = 0.0f;
		for(int i=0; i < n; i++)
			aux += u[i] * v[i];
		
		return aux;
		
	}
	
	
	
	
	void  denoising_block_to_vector(float **fpI, float *fpV, int i, int j, int iWin, int iWidth, int iChannels)
	{
		int kk = 0;
		for(int r=0; r < iWin; r++)
			for(int s= 0; s < iWin; s++)
			{
				int m = (j + r) * iWidth + i + s;
				for (int ii=0; ii < iChannels; ii++, kk++)			fpV[kk] = fpI[ii][m]; 
			}
	}
	
	
	
	void  denoising_block_to_vector(float *fpI, float *fpV, int i, int j, int iWin, int iWidth)
	{
		int kk = 0;
		for(int r=0; r < iWin; r++)
			for(int s= 0; s < iWin; s++, kk++)
			{
				int m = (j + r) * iWidth + i + s;
				fpV[kk] = fpI[m]; 
			}
	}
	
	
	
	
	
	
	
	
	
	
	void denoising_add_vector_to_block(float **fpO, float *fpM, float *fpM2, float *fpV, float multiplier, int ci, int cj, int iWin, int iWidth, int iChannels)
	{
		
		
		int kk = 0;
		for(int r=0; r < iWin; r++)
			for(int s=0; s < iWin; s++) 
			{	
				int m = (cj + r) * iWidth + ci + s;    
				
				for (int ic = 0; ic < iChannels; ic++, kk++) fpO[ic][m]   +=  multiplier * fpV[kk]; 
				
				if (fpM) fpM[m] += multiplier;
			}
		
		
		if (fpM2) fpM2[ cj * iWidth + ci]++;
	}
	
	
	
	void  denoising_add_vector_to_block(float *fpO, float *fpM, float *fpM2, float *fpV, float multiplier, int ci, int cj, int iWin, int iWidth)
	{
		
		
		int kk = 0;
		for(int r=0; r < iWin; r++)
			for(int s=0; s < iWin; s++, kk++) 
			{	
				int m = (cj + r) * iWidth + ci + s;    
				
				fpO[m]   +=  multiplier * fpV[kk]; 
				
				if (fpM) fpM[m] += multiplier;
			}
		
		if (fpM2) fpM2[ cj * iWidth + ci]++;
		
	}
	
	
	
	
	
	
	
	
	void denoising_dct_threshold(float *fpI, float *fpA, float *fpO, int iWidth, int iHeight, int iWin, float fRMult, float fSigma, int flagOverlapping, int flagWiener, int iSampling, int flagFlat, float valueFlat)
	{
		
		
		//! Pparameters
		int p = iWin * iWin;				// number of pixels per window		
		float fRMS = fRMult * fSigma;
		float fSigma2 = fSigma * fSigma; 
		float fRMS2 = fRMS * fRMS;
		
		float valueFlat2 = fSigma2 * valueFlat * valueFlat;
		
		//! mask: weight sum per pixel
		float *mask = new float[iWidth*iHeight];
		fpClear(mask,0.0f,iWidth*iHeight);
		
		
		//! clear output
		fpClear(fpO, 0.0f, iWidth*iHeight);
		
		
		//! dct basis
		float **V = new float*[p];
		for (int ii=0; ii < p; ii++) V[ii] = new float[p];
		
		int kk = 0;
		float fPi = 3.14159265f;
		for (int q0 = 0; q0 < iWin; q0++)
			for (int q1 = 0; q1 < iWin; q1++, kk++)
			{
				
				float aq0, aq1;
				
				if (q0 == 0)
					aq0 = sqrt(1.0 / (float) iWin);
				else				
					aq0 = sqrt(2.0 / (float) iWin);
				
				if (q1 == 0)
					aq1 = sqrt(1.0 / (float) iWin);
				else				
					aq1 = sqrt(2.0 / (float) iWin);
				
				
				int rr = 0;
				for (int m = 0; m < iWin; m++)
					for (int n=0; n < iWin; n++, rr++)
					{
						V[rr][kk] = aq0 * aq1 * cos( 0.5f * fPi * (2.0 * (float) m + 1.0) * (float) q0 / (float) iWin  ) * cos( 0.5f * fPi * (2.0 * (float) n + 1.0) * (float) q1 / (float) iWin  );
						
					}
				
				
			}
		
		
		
#pragma omp parallel shared(fpI, fpO)
		{
			
			
#pragma omp for schedule(dynamic) nowait 
			
			//! for each line
			for (int y = 0; y <= iHeight - iWin; y +=  iSampling)
			{
				
				
				//! memory for data matrix X
				float *X = new float[p];
				float *X0 = new float[p];
				float *XR = new float[p];
				
				
				//! memory for coefficients 
				float  *U = new float[p];
				float  *U0 = new float[p];
				float  *M = new float[p];
				
				
				
				//! for each pixel in current line
				for (int x = 0; x <= iWidth - iWin; x += iSampling)
				{	
					
					
					//! Build data vector X
					denoising_block_to_vector(fpI, X, x, y, iWin, iWidth);
					
					if (fpA)
						denoising_block_to_vector(fpA, X0, x, y, iWin, iWidth);
					
					
					
					//! Compute DCT coefficients
					for(int jj=0; jj < p; jj++)
					{
						
						float aux = 0.0; 
						for (int kk=0; kk <p; kk++) 
							aux +=  X[kk] * V[kk][jj]; 
						U[jj] =  aux; 
						
					}
					
					if (fpA)
					{
						
						for(int jj=0; jj < p; jj++)
						{
							
							float aux = 0.0; 
							for (int kk=0; kk <p; kk++) 
								aux +=  X0[kk] * V[kk][jj]; 
							U0[jj] =  aux; 
							
						}
						
						
					}
					
					
					// Compute standard deviation
					float mean = fpMean(X, p);
					float var = fpVar(X, p);
					
					float weight = 1.0;
					if (flagFlat && var < valueFlat2)
					{
						
						//! Reconstruct denoised patches
						for (int kk=0; kk < p; kk++)
							XR[kk] = mean;
						
						
						
					} else
					{
						//! Modify Coefficients
						if (!fpA)
						{
							
							for (int jj = 0; jj < p; jj++ ) 
							{
								
								if (fabsf(U[jj]) <  fRMS)  
								{
									M[jj] = 0.0;
								}
								else 
								{ 
									if (!flagWiener)  M[jj] = 1.0;
									else M[jj] = (U[jj]*U[jj] - fRMS2) / (U[jj]*U[jj]);
								}
								
								U[jj] *= M[jj];
							}
							
							
						}
						else 
						{
							
							for (int jj = 0; jj < p; jj++ ) 
							{
								
								if (!flagWiener)
								{
									if (fabsf(U0[jj]) < fRMS)  
									{
										M[jj] = 0.0;
									}
									else 
									{
										M[jj] = 1.0;
									}
									
								} else
								{
									M[jj] = (U0[jj]*U0[jj]) / (U0[jj] * U0[jj] +  fRMS2  /*fSigma2*/);
								}
								
								U[jj] *= M[jj];
								
								
							}
							
							
						}
						
						
						
						
						//! Compute weight for reducing Gibbs effect
						if (flagOverlapping == 2)
						{
							float energy = 0.0; 
							for(int jj = 0; jj < p; jj++) energy += M[jj] * M[jj];
							
							if (energy > 1.0f)
								weight = 1.0 / energy;
							else 
								weight = 1.0;
							
						}	
						
						
						
						
						//! Reconstruct denoised patches
						for (int kk=0; kk < p; kk++)
						{
							XR[kk] = scalar_product(U, V[kk],p);
						}
					}
					
					
					
					//! Put back denoised values into image
					if (flagOverlapping > 0)
						denoising_add_vector_to_block(fpO, mask, NULL, XR, weight, x , y, iWin, iWidth);
					else
					{
						fpO[(y + iWin / 2) * iWidth + (x + iWin / 2)] +=  XR[ (iWin/2) * iWin + (iWin/2)];
						mask[(y + iWin / 2) * iWidth + (x + iWin / 2)] += 1.0f;
						
					}
					
					
					
					
				}
				//! End of pixel x in each line
				
				
				
				delete[] X;
				delete[] XR;
				delete[] U;
				
				
			}
			//! End of each line
			
			
			
		}
		//! End of pragma 
		
		
		
		
		
		//! Normalize denoised values
		for(int ii=0; ii < iWidth*iHeight; ii++)
			if (mask[ii] > 0.0f) 
				fpO[ii] /= mask[ii];
			else 
				fpO[ii] = fpI[ii];
		
		
		
		delete[] mask;
		
		
		
		
	}
	
	
	
	
	
	
	
	
	
	void nlmeans_ipol(int iDWin,			// Half size of comparison window  
					  int iDBloc,		// Half size of research window
					  float fSigma,		// Noise parameter
					  float fFiltPar,	// Filtering parameter
					  float **fpI,		// Input
					  float **fpO,		// Output
					  int iChannels, int iWidth,int iHeight)
	{
		
		
		// length of each channel
		int iwxh = iWidth * iHeight;
		
		
		//  length of comparison window
		int ihwl = (2*iDWin+1);	
		int iwl = (2*iDWin+1) * (2*iDWin+1);
		int icwl = iChannels * iwl;
		
		
		
		// filtering parameter
		float fSigma2 = fSigma * fSigma;	
		float fH = fFiltPar * fSigma;
		float fH2 = fH * fH;
		
		// multiply by size of comparison window, since distances are not normalized
		fH2 *= (float) icwl;
		
		
		
		// tabulate exp(-x), faster than using directly function expf
		int iLutLength = (int) rintf((float) LUTMAX * (float) LUTPRECISION);
		float *fpLut = new float[iLutLength];
		wxFillExpLut(fpLut,iLutLength);
		
		
		
		
		// auxiliary variable
		// number of estimated values per pixel
		float *fpCount = new float[iwxh];
		fpClear(fpCount, 0.0f,iwxh);
		
		
		
		// clear output
		for (int ii=0; ii < iChannels; ii++) fpClear(fpO[ii], 0.0f, iwxh);
		
		
		
		
		// PROCESS STARTS
		// for each pixel (x,y)
#pragma omp parallel shared(fpI, fpO)
		{
			
			
#pragma omp for schedule(dynamic) nowait 
			
			for(int y=0; y < iHeight ; y++)
			{
				
				
				// auxiliary variable
				// denoised patch centered at a certain pixel
				float **fpODenoised = new float*[iChannels];
				for (int ii=0; ii < iChannels; ii++) fpODenoised[ii] = new float[iwl];
				
				
				
				
				for(int x=0 ; x < iWidth;  x++)
				{
					
					// reduce the size of the comparison window if we are near the boundary  
					int iDWin0 = MIN(iDWin,MIN(iWidth-1-x,MIN(iHeight-1-y,MIN(x,y))));
					
					
					// research zone depending on the boundary and the size of the window
					int imin=MAX(x-iDBloc,iDWin0);
					int jmin=MAX(y-iDBloc,iDWin0);
					
					int imax=MIN(x+iDBloc,iWidth-1-iDWin0);
					int jmax=MIN(y+iDBloc,iHeight-1-iDWin0);
					
					
					
					//  clear current denoised patch
					for (int ii=0; ii < iChannels; ii++) fpClear(fpODenoised[ii], 0.0f, iwl);
					
					
					
					// maximum of weights for central pixel in case exponential strategy is used
					float fMaxWeight = 0.0f;
					
					// sum of weights
					float fTotalWeight = 0.0f;
					
					
					
					for(int j=jmin; j <= jmax; j++)
						for(int i=imin ; i <= imax; i++) 
							if (i!=x || j!=y)
							{
								
                                //! Attention this functions divides by the number of channels
                                //! This is not the case for the one in IPOL
								float fDif = fiL2FloatDist(fpI,fpI,x,y,i,j,iDWin0, iDWin0, iChannels,iWidth,iWidth);

								//! IPOL command
                                // fDif = MAX(fDif - 2.0f * (float) icwl *  fSigma2, 0.0f);

								//! Current commmand
                                fDif = MAX(fDif - 2.0f * (float) iwl *  fSigma2, 0.0f);
                                
                                fDif = fDif / fH2;	
								
                                
								float fWeight = wxSLUT(fDif,fpLut);
								
								if (fWeight > fMaxWeight) fMaxWeight = fWeight;
								
								fTotalWeight += fWeight;	
								
								
								for(int is=-iDWin0; is <=iDWin0; is++)
								{	
									int aiindex = (iDWin+is) * ihwl + iDWin;
									int ail = (j+is)*iWidth+i;
									
									for(int ir=-iDWin0; ir <= iDWin0; ir++)
									{
										
										int iindex = aiindex + ir;
										int il= ail +ir;
										
										for(int ii=0; ii < iChannels; ii++)
											fpODenoised[ii][iindex] += fWeight * fpI[ii][il];
										
									}
									
								}
								
								
							} 
					
					
					
					
					// current point (x,y) with fMaxWeight
					
					for(int is=-iDWin0; is <=iDWin0; is++)
					{
						int aiindex = (iDWin+is) * ihwl + iDWin;
						int ail=(y+is)*iWidth+x;
						
						for(int ir=-iDWin0; ir <= iDWin0; ir++)
						{
							
							int iindex = aiindex + ir;
							int il=ail+ir;
							
							for(int ii=0; ii < iChannels; ii++)
								fpODenoised[ii][iindex] += fMaxWeight * fpI[ii][il];
							
						}
					}
					
					
					
					fTotalWeight += fMaxWeight;	
					
					
					
					
					
					
					
					// normalize average value
					// else case treated as fpCount[y*width+x] = 0
					if (fTotalWeight > fTiny) 
					{
						
						
						
						for(int is=-iDWin0; is <=iDWin0; is++)
						{
							int aiindex = (iDWin+is) * ihwl + iDWin;
							int ail=(y+is)*iWidth+x;
							
							for(int ir=-iDWin0; ir <= iDWin0; ir++)
							{
								int iindex = aiindex + ir;
								int il=ail+ ir;
								
								#pragma omp atomic
                                fpCount[il]++;
								
								for(int ii=0; ii < iChannels; ii++)
								{
									#pragma omp atomic
                                    fpO[ii][il] += fpODenoised[ii][iindex] / fTotalWeight; 
									
								}
								
							}
						}
						
						
					} 
					
					
					
					
					
					
				}
				
			}
			
			
			
			
			
			
			
			
		}
		
		
		
		
		for(int ii=0; ii < iwxh;ii++)
			if (fpCount[ii]>0.0)
			{
				for (int jj=0; jj < iChannels; jj++)  fpO[jj][ii] /= fpCount[ii];
				
			}	else 
			{
				
				for (int jj=0; jj < iChannels; jj++)  fpO[jj][ii] = fpI[jj][ii];
			}
		
		
		
		
		
	}
	
	
	
	
	
    
    
	
	float compute_knn_mean_std_float(int k, float *omean, float *ostd, float *dist, int n)
	{
		
		if (n > k)
		{
			float *bdist = new float[k];
			
			fpClear(bdist, 10000000000.0f, k);
			for(int ikt = 0; ikt < n; ikt++) 
			{
				
				float dif = dist[ikt];
				
				
				if (dif < bdist[k-1]) 
					
				{
					int counter;
					for(counter = k - 1; counter>0 && bdist[counter-1] > dif; counter--){bdist[counter] = bdist[counter-1];}
					
					bdist[counter] = dif;
				}
				
				
				
			}
			
			
			
			
			*omean = (float) fpMean(bdist, k);
			*ostd = (float) sqrt(fpVar(bdist, k));
			return bdist[k-1];
			
		}
		else
		{
			
			
			*omean = (float) fpMean(dist, n);
			*ostd = (float) sqrt(fpVar(dist, n));
			return fpMax(dist, NULL, n);
			
			
		}
		
		
		
		
	}
	
    
    
    
	
	
	
	void nlmeans(int iDWin,			// Half size of comparison window  
				 int iDBloc,		// Half size of research window
				 int ifOverlap,		// IF not zero THEN block synthesized and average at intersections
				 int ifThreshold,	// IF not zero THEN threshold function ELSE  exponential function
				 float * fpKernel,  int iKerWidth, // IF fpKernel != NULL THEN weighted window distance
				 int iSampling,		// Sampling 
                 int iK,            // use first kNN
				 float fSigma,		// Noise parameter
				 float fFiltPar,	// Filtering parameter
				 float **fpI,		// Input
				 float **fpA,		// Auxiliary input for distance computation
				 float **fpO,		// Output
				 int iChannels, int iWidth,int iHeight)
	{
		
		
		// length of each channel
		int iwxh = iWidth * iHeight;
		
		// length of comparison window
		int ihwl = (2*iDWin+1);	
		int iwl = (2*iDWin+1) * (2*iDWin+1);
		int icwl = iChannels * iwl;
		
		// length of research block
		int ibl = (2*iDBloc+1) * (2*iDBloc+1);
		
		
		// std of squared distances of patches of size icwl
		float fSigma2 = fSigma * fSigma;	
		float fVariance = (8.0f * fSigma2 * fSigma2) / (float) (icwl);
		float fStd = sqrtf(fVariance);
		
		
		// threshold  case		
		/////		weight => 1 		if dist < t1
		////		(t2-x)/(t2-t1)	if t1<dist<t2
		////		0		if t>t2
		float	t1 = 2.0f * fSigma2 ;							// mean
		float	t2 = 2.0f * fSigma2 + fFiltPar * fStd;   		// mean + m * std
		
		// flat distance is not normalized
		if (!fpKernel)
		{	
			t1 *= (float) icwl;			// mean
			t2 *= (float) icwl;   		// mean + m * std
		}
		
		float t2mt1 = t2 - t1 ;								// useful for weight computation
		
		
		// exponential case
		int iLutLength = (int) rintf((float) LUTMAX * (float) LUTPRECISION);
		float *fpLut = new float[iLutLength];
		wxFillExpLut(fpLut,iLutLength);
		
		
		
		float fH = fFiltPar * fSigma;
		float fH2 = fH * fH;
		
		// flat distance is not normalized
		if (!fpKernel)
		{
			fH2 *= (float) icwl;
		}
		
		// normalize kernel
		if (fpKernel)
		{
			
			iDWin = (iKerWidth -1) / 2;    
			
			float fSum = 0.0f;
			for (int ii=0; ii < iKerWidth * iKerWidth; ii++) fSum += fpKernel[ii];
			
			
			if (fSum < fTiny)  { printf("kernel cannot be normalized"); exit(-1);}
			
			for (int ii=0; ii < iKerWidth * iKerWidth; ii++)  fpKernel[ii] /= fSum;
			
		}
		
		
		
		
		
		// auxiliary variable
		// number of estimated values per pixel
		float *fpCount = new float[iwxh];
		fpClear(fpCount, 0.0f,iwxh);
		
		
		
		// clear output
		for (int ii=0; ii < iChannels; ii++) fpClear(fpO[ii], 0.0f, iwxh);
		
		
		
		
		// PROCESS STARTS
		// for each pixel (x,y)
#pragma omp parallel shared(fpI, fpO)
		{
			
			
#pragma omp for schedule(dynamic) nowait 
			
			for(int y=0; y < iHeight ; y += iSampling)
			{
				
				// auxiliary variable
				// denoised value at a certain pixel
				float *fpDenoised = new float[iChannels];
				
				
				// auxiliary variable
				// denoised patch centered at a certain pixel if ifOverlap activated
				float **fpODenoised = new float*[iChannels];
				for (int ii=0; ii < iChannels; ii++) fpODenoised[ii] = new float[iwl];
				
				// store patch distances
				float *fpDist = new float[ibl];
				
				
				
				for(int x=0 ; x < iWidth; x += iSampling)
				{
					
					int ixy = y*iWidth+x;
					
					// reduce the size of the comparison window if we are near the boundary  
					int iDWin0 = MIN(iDWin,MIN(iWidth-1-x,MIN(iHeight-1-y,MIN(x,y))));
					
					// research zone depending on the boundary and the size of the window
					int imin=MAX(x-iDBloc,iDWin0);
					int jmin=MAX(y-iDBloc,iDWin0);
					
					int imax=MIN(x+iDBloc,iWidth-1-iDWin0);
					int jmax=MIN(y+iDBloc,iHeight-1-iDWin0);
					
					
					
					
					
					//  clear current denoised value
					fpClear(fpDenoised, 0.0f, iChannels);
					
					//  clear current denoised patch
					for (int ii=0; ii < iChannels; ii++) fpClear(fpODenoised[ii], 0.0f, iwl);
					
					
					//  compute l2 distances
					int ikt = 0;
					for(int j=jmin; j <= jmax; j++)
						for(int i=imin ; i <= imax; i++, ikt++) 
						{
							
							
							float fDif;
							
							if (fpA)	// if auxiliary image for distance computation
							{
								if (fpKernel && iDWin == iDWin0)
									fDif = fiL2FloatWDist(fpA,fpA,x,y,i,j,iDWin0,iDWin0,fpKernel,iChannels,iWidth,iWidth);
								else 
									fDif = fiL2FloatDist(fpA,fpA,x,y,i,j,iDWin0, iDWin0, iChannels,iWidth,iWidth);
								
							} else 
							{
								
								if (fpKernel && iDWin == iDWin0) 
									fDif = fiL2FloatWDist(fpI,fpI,x,y,i,j,iDWin0,iDWin0,fpKernel,iChannels,iWidth,iWidth);
								else 
									fDif = fiL2FloatDist(fpI,fpI,x,y,i,j,iDWin0,iDWin0,iChannels,iWidth,iWidth);
								
							}
							
							fpDist[ikt] = fDif;
							
						}
					
                    
                    float iomeank, iostdk;
                    float fDifK;
                    fDifK =   compute_knn_mean_std_float(iK, &iomeank, &iostdk, fpDist, ikt);
                    
					
					// maximum of weights for central pixel in case exponential strategy is used
					float fMaxWeight = 0.0f;
					
					// sum of weights
					float fTotalWeight = 0.0f;
					
					
					
					ikt = 0;
					for(int j=jmin; j <= jmax; j++)
						for(int i=imin ; i <= imax; i++, ikt++) 
						{
							int iij = j*iWidth+i;
							
							float fDif = fpDist[ikt];
							float fWeight = 0.0f;
							
							
							
							// exponential case and not current pixel (x,y)
							if (!ifThreshold && !iK && (i!=x || j!=y))
							{
								
								//fDif = MAX(fDif - 2.0f * (float) icwl *  fSigma2, 0.0f);
								fDif = MAX(fDif, 0.0f);
								
								
								if (fpKernel && iDWin0 != iDWin)
									fDif /= (float) icwl;
								
								fDif = fDif / fH2;	
								fWeight = wxSLUT(fDif,fpLut);
								//fWeight =expf(-fDif);
								
								if (fWeight > fMaxWeight) fMaxWeight = fWeight;
								
								fTotalWeight += fWeight;	
								
								
							} else 	if (ifThreshold && !iK && fDif < t2)	
							{					
								
								if (fDif < t1) fWeight = 1.0f;
								else fWeight = (t2 - fDif) / t2mt1;
								
								fTotalWeight += fWeight;	
								
							} else if (iK && fDif < fDifK)
							{
                                
                            	fWeight = 1.0f;
								fTotalWeight += fWeight;	
                                
                            }
							
							
							// if using threshold and fDif < t2
							// or exponential and not current point (x,y)
							if ((!ifThreshold && (i!=x || j!=y)) || (ifThreshold && fDif < t2))
							{
								
								if (!ifOverlap)
								{
									for (int ii=0; ii < iChannels; ii++)  fpDenoised[ii] +=  fWeight * fpI[ii][iij];
									
								} else {
									
									for(int is=-iDWin0; is <=iDWin0; is++)
									{	
										int aiindex = (iDWin+is) * ihwl + iDWin;
										int ail = (j+is)*iWidth+i;
										
										for(int ir=-iDWin0; ir <= iDWin0; ir++)
										{
											
											int iindex = aiindex + ir;
											int il= ail +ir;
											
											for(int ii=0; ii < iChannels; ii++)
												fpODenoised[ii][iindex] += fWeight * fpI[ii][il];
											
										}
										
									}
									
								}
								
								
							}
							
							
						}
					
					
					
					
					
					// exponential and current point (x,y)
					if (!ifThreshold)
					{
						
						if (!ifOverlap)
						{
							for (int ii=0; ii < iChannels; ii++)  fpDenoised[ii] +=  fMaxWeight * fpI[ii][ixy];
							
						} else {
							
							for(int is=-iDWin0; is <=iDWin0; is++)
							{
								int aiindex = (iDWin+is) * ihwl + iDWin;
								int ail=(y+is)*iWidth+x;
								
								for(int ir=-iDWin0; ir <= iDWin0; ir++)
								{
									
									int iindex = aiindex + ir;
									int il=ail+ir;
									
									for(int ii=0; ii < iChannels; ii++)
										fpODenoised[ii][iindex] += fMaxWeight * fpI[ii][il];
									
								}
							}
							
						}
						
						fTotalWeight += fMaxWeight;	
						
						
					}
					
					
					
					
					// normalize average value
					// else case treated as fpCount[y*width+x] = 0
					if (fTotalWeight > fTiny) 
					{
						
						if (!ifOverlap)
						{
							for (int ii=0; ii < iChannels; ii++)  fpO[ii][y*iWidth+x] += fpDenoised[ii] / fTotalWeight;
							fpCount[y*iWidth+x]++;
							
						} else {
							
							for(int is=-iDWin0; is <=iDWin0; is++)
							{
								int aiindex = (iDWin+is) * ihwl + iDWin;
								int ail=(y+is)*iWidth+x;
								
								for(int ir=-iDWin0; ir <= iDWin0; ir++)
								{
									int iindex = aiindex + ir;
									int il=ail+ ir;
									
									fpCount[il]++;
									
									for(int ii=0; ii < iChannels; ii++)
									{
										fpO[ii][il] += fpODenoised[ii][iindex] / fTotalWeight; 
										
									}
									
								}
							}
						}
						
					} 
					
					
					
					
					
					
				}
				
			}
			
			
		}
		
		
		
		
		for(int ii=0; ii < iwxh;ii++)
			if (fpCount[ii]>0.0)
			{
				for (int jj=0; jj < iChannels; jj++)  fpO[jj][ii] /= fpCount[ii];
				
			}	else 
			{
				
				for (int jj=0; jj < iChannels; jj++)  fpO[jj][ii] = fpI[jj][ii];
			}
		
		
		
	}
	
	
	
	
	
	
	
	
	
	
	
	
	float compute_knn_mean_std(int k, double *omean, double *ostd, double *dist, int n)
	{
		
		if (n > k)
		{
			double *bdist = new double[k];
			
			dpClear(bdist, 10000000000.0f, k);
			for(int ikt = 0; ikt < n; ikt++) 
			{
				
				float dif = dist[ikt];
				
				
				if (dif < bdist[k-1]) 
					
				{
					int counter;
					for(counter = k - 1; counter>0 && bdist[counter-1] > dif; counter--){bdist[counter] = bdist[counter-1];}
					
					bdist[counter] = dif;
				}
				
				
				
			}
			
			
			
			
			*omean = (double) dpMean(bdist, k);
			*ostd = (double) sqrt(dpVar(bdist, k));
			return bdist[k-1];
			
		}
		else
		{
			
			
			*omean = (float) dpMean(dist, n);
			*ostd = (float) sqrt(dpVar(dist, n));
			return dpMax(dist, NULL, n);
			
			
		}
		
		
		
		
	}
	
	
	
	
	
	
	
	/*
     void wiener_coefficients(float **U, float **M, float rsigma2, int p, int npts)
     {
     
     float fAux;
     for(int jj=0; jj < p; jj++)
     for(int ii=0; ii < npts; ii++)		
     {
     
     fAux =  U[ii][jj] * U[ii][jj];
     
     if (fAux < rsigma2)
     M[ii][jj] = 0.0f;
     else 
     M[ii][jj] = (fAux - rsigma2) / fAux;
     }
     }
     */
	
	
	void wiener_oracle_coefficients(double **U, double **M, double rsigma2, int p, int npts)
	{
		
		double fAux;
		for(int jj=0; jj < p; jj++)
			for(int ii=0; ii < npts; ii++)		
			{
				fAux =  U[ii][jj] * U[ii][jj];
				
				M[ii][jj] = fAux / (fAux + rsigma2);
			}
		
	}
	
	
	
	void hard_coefficients(double **U, double **M, double rsigma2, int p, int npts)
	{
		double fAux;
		for(int jj=0; jj < p; jj++)
			for(int ii=0; ii < npts; ii++)
			{
				fAux =  U[ii][jj] * U[ii][jj];
				
				if (fAux < rsigma2)
					M[ii][jj] = 0.0f;
				else
					M[ii][jj] = 1.0f;
			}
	}
	
	
	
	
	void wiener_principal_values(double *S, double **M, double rsigma2, int p, int npts)
	{
		
		for(int jj=0; jj < p; jj++)
			for(int ii=0; ii < npts; ii++)		
				if (S[jj] < rsigma2)
					M[ii][jj] = 0.0f;
				else 
					M[ii][jj] = (S[jj] - rsigma2) / S[jj];
		
		
	}
	
	
	/*
     void wiener_oracle_principal_values(float *S, float **M, float rsigma2, int p, int npts)
     {
     
     
     for(int jj=0; jj < p; jj++)
     for(int ii=0; ii < npts; ii++)		
     M[ii][jj] = S[jj] / (S[jj] + rsigma2);
     
     
     }
     */
	
	
	
	void hard_principal_values(double *S, double **M, double rsigma2, int p, int npts)
	{
		for(int jj=0; jj < p; jj++)
			for(int ii=0; ii < npts; ii++)
				if (S[jj] < rsigma2)
					M[ii][jj] = 0.0f;
				else
					M[ii][jj] = 1.0f;
	}
	
	
	
	
	
	
	
	/*
     void hard_principal_adaptive_values(float *S, float **M, float *cest, float sigma,float fMult, int p,int npts)
     {
     
     //// Order singular values
     float *Sorder = new float[p];
     int *oindex = new int[p];
     
     for(int ii=0; ii < p; ii++)  Sorder[ii] = S[ii];
     
     for(int ii = 0; ii < p; ii++)  oindex[ii] = ii;
     
     order_decreasing(Sorder,oindex,p);		// In[j]: position of j-ordered element in original vector
     
     
     /// Compute inverse search indexos
     int *oindex2 = new int[p];	
     for(int ii=0; ii < p; ii++)  oindex2[oindex[ii]] = ii;
     
     
     
     float *curve = new float[p];
     fpCopy(cest, curve, p);
     
     
     float sigma2 = sigma * sigma;
     for (int ii=0; ii < p; ii++) 
     if (curve[ii] < sigma2 )  curve[ii] = sigma2; 
     
     
     for(int jj=0; jj < p; jj++)
     for(int ii=0; ii < npts; ii++)		
     if (S[jj] < fMult * fMult * curve[oindex2[jj]])
     M[ii][jj] = 0.0f;
     else 
     M[ii][jj] = 1.0f;
     
     
     }
     */	
	
    /*
     void wiener_principal_adaptive_values(float *S, float **M, float *cest, float sigma,float fMult, int p,int npts)
     {
     
     
     //// Order singular values
     float *Sorder = new float[p];
     int *oindex = new int[p];
     
     for(int ii=0; ii < p; ii++)  Sorder[ii] = S[ii];
     
     for(int ii = 0; ii < p; ii++)  oindex[ii] = ii;
     
     order_decreasing(Sorder,oindex,p);		// In[j]: position of j-ordered element in original vector
     
     
     /// Compute inverse search indexos
     int *oindex2 = new int[p];	
     for(int ii=0; ii < p; ii++)  oindex2[oindex[ii]] = ii;
     
     
     
     float *curve = new float[p];
     fpCopy(cest, curve, p);
     
     
     float sigma2 = sigma * sigma;
     for (int ii=0; ii < p; ii++) 
     if (curve[ii] < sigma2 )  curve[ii] = sigma2; 
     
     
     for(int jj=0; jj < p; jj++)
     for(int ii=0; ii < npts; ii++)		
     if (S[jj] < fMult * fMult * curve[oindex2[jj]])
     M[ii][jj] = 0.0f;
     else 
     M[ii][jj] = (S[jj] - fMult * fMult * curve[oindex2[jj]] ) / S[jj];
     
     
     }
     */	
	
	
	
    /*	
     void  estimate_noise_pca_variances(int d, int n, float sigma, float delta, int *osize, float *curve)
     {
     
     // ///////  d 			dimension
     // ///////  n    		samples
     // ///////  sigma		noise standard deviation
     // //  y = n / d;
     // //  a = a(y) =  sigma^2 * (1 − y^1/2)^2
     // //  b = b(y) =  sigma^2 * (1 + y^1/2)^2 
     // //  p_{y,sigma} (x) =  (1 / 2 pi y x sigma^2) sqrt((b-x) (a-x)) 	if a <= x <= b 
     // //			0						otherwise
     // //
     // //  p_{y, sigma} (x) =  p_{y,1} (x / sigma^2) / sigma^2
     // //
     // //  F_{sigma}(eta) = F_{1} (eta / sigma^2)
     
     int N = MIN(n,d);
     
     float y = (float) d / (float) n;
     float x;
     
     float sigma2 = sigma*sigma;
     float a = 1.0 - sqrt(y);
     a = sigma2 * a * a;  // Can be a*a > b*b? No because, then |1-sqrt(y)| > |1+sqrt(y)| 
     
     float b = 1.0 + sqrt(y);
     b = sigma2 * b * b;
     
     int isize = (int) ceilf( b / delta) + 2*11; // We begin at x=-10*delta .. b + 10*delta 
     *osize = isize;
     
     float *cdf = new float[isize];
     
     
     float normal = 2.0f * 3.14159f * y * sigma2;
     //printf("y: %f a: %f b: %f\n", y , a , b);
     
     if (y < 1) 
     {
     
     float cumulative = 0.0;
     x = -10.0f * delta;
     for(int i = 0; i < isize; i++, x+=delta)
     {
     
     float value;
     
     if (a<=x && x<=b) value =  delta * sqrtf( (b-x)*(x-a)) / (x * normal);
     else value = 0.0;
     
     cumulative += value;
     
     cdf[i] =  (float) N * (1.0 - cumulative); 
     //printf("%f: %f \n", x, cdf[i]);
     }	
     
     
     }  else{
     
     
     float cumulative = 1.0  -  1.0 / y;
     
     cdf[0] =  (float) d *   (1.0 - cumulative); 
     //printf("cumulative: %f n: %f cdf: %f \n", cumulative, (float) n, cdf[0]);
     
     x = delta;
     for(int i = 1; i < isize; i++,  x+=delta)
     {
     float value;
     if (a<=x && x<=b) value =  delta * sqrt( (b-x)*(x-a)) / (x * normal);
     else value = 0.0;
     
     cumulative +=   value;
     
     cdf[i] =  (float) d * (1.0 - cumulative); 
     }	
     
     
     } 
     
     
     float minpositiu = (float)(N + 1);
     for (int i=0; i < isize; i++) {
     
     if (cdf[i] < minpositiu && cdf[i] >= 0.0)  minpositiu = cdf[i];
     
     }
     
     
     int nvalue = N;	
     fpClear(curve, 0.0, N);
     x = -10.0 * delta;
     for(int i=0; nvalue > 0 && i < isize-1 ;i++, x += delta)
     {
     
     
     float ceps = 0.01;
     
     if (nvalue == N && cdf[i] > (float) N - ceps && cdf[i] != cdf[i+1])
     {
     curve[nvalue-1] = x; 
     nvalue--;
     
     } 
     // else if (nvalue == 0 && cdf[i] < minpositiu + ceps && cdf[i] != cdf[i+1])
     // {
     
     // curve[nvalue] = x; 
     // nvalue--;
     
     // } 
     else if (nvalue != N && cdf[i] >= (float) nvalue - ceps && cdf[i] < (float) nvalue + ceps && cdf[i] != cdf[i+1])
     {
     
     curve[nvalue-1] = x; 
     nvalue--;
     
     }
     
     }
     
     
     delete[] cdf;
     
     
     }
     */
	
	
	
	
	
	
	/*
     void wiener_filtering(float **fpI, float **fpA, float **fpO, float *fpSPC,
     char * option, int ifCopying,  int ifAOverlap, 
     int iDWin, int iDBloc,
     float fSigma, int idisttype,
     int iSampling, int ifK, 
     float fMult, float fRMult, int iChannels, int iWidth, int iHeight)
     {
     
     
     
     /// Parameters
     int n = (iWidth - 2 * iDWin - 1) * (iHeight - 2*iDWin - 1);
     int p = (2*iDWin + 1) * (2*iDWin + 1) * iChannels;	
     int bsize = (2*iDBloc+1) * (2*iDBloc+1);
     
     
     
     
     /// Print info 
     if (0)
     {
     printf("... b: %d x %d  w: %d x %d\n", 2*iDBloc+1, 2*iDBloc+1, 2*iDWin+1,2*iDWin+1);
     printf("... n: %d  p: %d\n", n, p);
     printf("... sampling: %d\n", iSampling);
     printf("... koption: %d\n", ifK);
     printf("... mult: %f  rmult: %f  sigma: %f\n", fMult, fRMult,  fSigma);
     printf("... type: %s\n", option);
     printf("... adaptive overlapping: %d \n", ifAOverlap);
     printf("... copying: %d \n",ifCopying);
     
     if (fpA)  printf("... using secondary image\n");
     }
     
     if (0)
     {
     if (idisttype == 1)
     printf("... using fixed distance threshold %f\n", fMult);
     
     else if (idisttype == 11)
     printf("... using maximum of fixed distance threshold %f and %d nn values \n", fMult, ifK);
     
     else if (idisttype == 2)
     printf("... using KNN with %d\n", ifK);
     
     else if (idisttype == 3)
     printf("... using fixed distance threshold %f with at maximum %d neighbours (bm3d strategy)\n", fMult, ifK);
     
     else if (idisttype == 4)
     printf("... using chi average plus %f times the chi std\n", fMult);
     
     else if (idisttype == 5)
     printf("... using chi average plus %f times the chi std with at maximum %d neighbours\n", fMult, ifK);
     
     else if (idisttype == 6)
     printf("... using kNN average plus %f times the KNN std with %d neighbours\n", fMult, ifK);
     
     else if (idisttype == 7)
     printf("... using the maximum of average plus %f times the std, of the chi square and %d nn (current nlpca strategy)\n", fMult, ifK);			
     
     else if (idisttype == 8)
     printf("... using the maximum of average plus %f times the std of noise and the %d NN\n", fMult, ifK);			
     
     
     
     }
     
     
     
     
     /// Mask: weight sum per pixel
     float *mask = new float[iWidth*iHeight];
     fpClear(mask,0.0f,iWidth*iHeight);
     
     
     
     /// Mask2: treated pixels being central pixels of the window 
     ///        or copied by copy paste strategy being center of the window
     float *mask2 = new float[iWidth*iHeight];
     fpClear(mask2,0.0f,iWidth*iHeight);
     
     
     
     
     // clear output
     for (int ii=0; ii < iChannels; ii++) fpClear(fpO[ii], 0.0f, iWidth*iHeight);
     
     
     #pragma omp parallel shared(fpI, fpO)
     {
     
     
     #pragma omp for schedule(dynamic) nowait 
     
     
     for (int y = iDWin; y < iHeight - iDWin - 1; y+= iSampling)
     {
     
     
     
     
     /// Memory for PCA (memory allocated in here in order to allow the use of parallel for)
     float **X2 = allocate_float_matrix(bsize, p);
     float **XR2 = allocate_float_matrix(bsize, p);
     float *baricenter = new float[p];
     float *S=NULL, **V=NULL, **U=NULL, **M=NULL;
     
     
     int flag_computing_pca=0;
     if (strcmp(option,"w") == 0  ||  strcmp(option,"h") == 0 ||  strcmp(option,"ch") == 0 ||  strcmp(option,"cw") == 0  ||  strcmp(option,"ah") == 0 ||   strcmp(option,"aw") == 0 )
     flag_computing_pca = 1;
     
     
     if (( strcmp(option,"ah") == 0 ||  strcmp(option,"aw") == 0 ) && fpSPC == NULL) {printf("... error :: list of variances necessary for adaptive thresholding (t=ah or t=aw )\n"); exit(-1);}
     
     
     
     if (flag_computing_pca)
     {
     S = new float[p];
     V = allocate_float_matrix(p, p);
     U = allocate_float_matrix(bsize, p);
     M = allocate_float_matrix(bsize, p);
     }
     
     
     
     ///  Memory for the computation of PCA with a secondary image
     float **X02 = NULL, **U02 = NULL;
     float *baricenter0 = NULL;
     
     if (fpA && flag_computing_pca)
     {
     U02 = allocate_float_matrix(bsize, p);
     X02 = allocate_float_matrix(bsize, p);
     baricenter0 = new float[p];
     }
     
     
     
     
     /// Vector of squared distances 
     float *tdistances = new float[bsize];
     
     
     
     /// Other variables
     int scurrent = 0;							// position of reference pixel in the X2 table
     int *indexos = new int[n];					// indexos[k]: coordinates of block in k-entry of X2
     int *indexos2 = new int[n];					// indexos2[k]: position in neighborhood of k-entry of X2
     
     
     
     
     
     for (int x = iDWin; x < iWidth - iDWin - 1; x+=iSampling)
     if ( (ifCopying == 0) || (ifCopying == 1) || (ifCopying == 2 && mask2[y*iWidth+x] == 0.0f))		// c = 0,1 all blocks denoised as reference blocks
     {																								// c = 2   blocks can be denoised as neighborhood of a reference block
     
     
     
     /// Learning zone
     int imin=MAX(x-iDBloc,iDWin);
     int jmin=MAX(y-iDBloc,iDWin);
     int imax=MIN(x+iDBloc,iWidth-1-iDWin);
     int jmax=MIN(y+iDBloc,iHeight-1-iDWin);
     
     
     
     /// Computing distances
     fpClear(tdistances,0.0f,bsize);
     int ikt = 0;
     for(int j=jmin;j<=jmax;j++)
     for(int i=imin;i<=imax;i++,ikt++) 
     {
     
     float fDif;	
     
     
     if (fpA) 
     fDif  = fiL2FloatDist(fpA,fpA,x,y,i,j,iDWin,iChannels, iWidth,iWidth);  
     
     else 
     fDif  = fiL2FloatDist(fpI,fpI,x,y,i,j,iDWin,iChannels, iWidth,iWidth);  
     
     
     
     
     
     
     tdistances[ikt] = fDif;
     }
     
     
     
     //!  Set threshold for patches
     //! 1. Use fixed value
     //! 2. Use KNN
     //! 3. Use fixed value with at maximum k neighbours
     //! 4. Use average + m * std
     //! 5. Use average + m * std with at maximum k neighbours
     //! 6. Use average + m * std being average and std of the first KNN
     //! 7. Use the maximum of 4 and 5
     
     float iomeank, iostdk, difk;
     difk = compute_knn_mean_std(ifK, &iomeank, &iostdk, tdistances, ikt);
     float fSigma2 = fSigma * fSigma;
     
     float pthresh=0.0;
     if (idisttype == 1)
     pthresh = fMult * fMult * (float) p;
     
     else if (idisttype == 11)
     {
     pthresh = fMult * fMult * (float) p;
     pthresh = MAX(pthresh, difk);
     
     }else if (idisttype == 2)
     pthresh = difk;
     
     else if (idisttype == 3)
     pthresh = MIN(fMult * fMult * (float) p , difk);
     
     else if (idisttype == 4)
     {
     float average = 2.0f * fSigma2 * (float) p;
     float variance = (8.0f * fSigma2 * fSigma2) / (float) (p);
     float std = (float) p * (float) sqrt(variance);
     
     pthresh = (float) average + fMult * std; 
     }		
     else if (idisttype == 5)
     {
     float average = 2.0f * fSigma2 * (float) p;
     float variance = (8.0f * fSigma2 * fSigma2) / (float) (p);
     float std = (float) p * (float) sqrt(variance);
     
     pthresh = MIN((float) average + fMult * std, difk);
     
     }		
     else if (idisttype == 6)
     {
     pthresh = iomeank + fMult * iostdk;
     
     }						
     else if (idisttype == 7)
     {
     float average = 2.0f * fSigma2 * (float) p;
     float variance = (8.0f * fSigma2 * fSigma2) / (float) (p);
     float std = (float) p * (float) sqrt(variance);
     
     pthresh = MAX(average + fMult * std, iomeank + fMult * iostdk );
     
     } else if (idisttype == 8)
     {
     
     float average = 2.0f * fSigma2 * (float) p;
     float variance = (8.0f * fSigma2 * fSigma2) / (float) (p);
     float std = (float) p * (float) sqrt(variance);
     
     pthresh = MAX(average + fMult * std, difk);
     
     }
     
     
     
     /// Selecting similar patches
     int npts=0;
     ikt = 0;
     for(int j=jmin;j<=jmax;j++)
     for(int i=imin;i<=imax;i++,ikt++) 
     {
     
     float fDif = tdistances[ikt];
     if (fDif < pthresh)
     {
     
     
     // Fill X2 with noisy patch
     int kk = 0;
     for (int ii=0; ii < iChannels; ii++) 
     {
     for(int r=-iDWin; r <= iDWin; r++)
     for(int s=-iDWin; s <= iDWin; s++,kk++)
     { int m = (j + r) * iWidth + i + s;  X2[npts][kk] = fpI[ii][m]; }
     
     }
     
     
     
     // Fill X2 with additional patch if additional image selected
     if (fpA &&  flag_computing_pca)
     {
     kk=0;
     for (int ii=0; ii < iChannels; ii++) 
     {
     for(int r=-iDWin; r <= iDWin; r++)
     for(int s=-iDWin; s <= iDWin; s++,kk++)   { int m = (j + r) * iWidth + i + s; X02[npts][kk] = fpA[ii][m]; }
     }
     }	
     
     
     indexos[npts] = j*iWidth+i;					
     indexos2[npts] = ikt;	
     if (y == j && x == i) scurrent = npts;
     
     npts++;
     }
     
     }	
     
     
     
     
     
     // Case 
     float *weight = new float[npts];
     fpClear(weight, 1.0f, npts);
     
     if (strcmp(option,"b1") == 0)													// l1 baricenter
     {
     //l1_baricenter_mauricio(X2, baricenter, npts, p);
     
     //for (int ii=0; ii < npts; ii++) fpCopy(baricenter, XR2[ii], p);
     
     } else if (strcmp(option,"b2") == 0 ||  (flag_computing_pca && npts < MNPCA))	// l2 baricenter
     {
     
     l2_baricenter(X2, baricenter, npts, p);
     for (int ii=0; ii < npts; ii++) fpCopy(baricenter, XR2[ii], p);				
     
     } else																			// PCA
     {
     
     /// Compute baricenter and center data
     pca_center_data(X2, baricenter, npts, p);
     compute_pca_svd(X2, S, V, U, npts, p);
     
     
     if (fpA && flag_computing_pca)
     { 
     pca_center_data(X02, baricenter0, npts, p); 
     compute_pca_svd(X02, S, V, U02, npts, p);
     
     // Compute U coefficients with PCA basis V
     for(int ii=0; ii < npts; ii++)
     { 
     for(int jj=0; jj < p; jj++)
     {
     float aux = 0.0; 
     for (int kk=0; kk <p; kk++) 
     aux += X2[ii][kk] * V[kk][jj]; U[ii][jj] =  aux; 
     }
     }
     
     }
     
     
     
     /// Modify Coefficients
     float fRSigma2 = fRMult * fRMult * fSigma2;
     
     if (strcmp(option,"w") == 0 && !fpA)  
     wiener_principal_values(S,M,fRSigma2,p,npts);
     
     else if (strcmp(option,"h") == 0 && !fpA) 
     hard_principal_values(S,M,fRSigma2,p,npts);
     
     else if (strcmp(option,"ch") == 0 && !fpA)  
     hard_coefficients(U,M,fRSigma2,p,npts);
     
     else if (strcmp(option,"cw") == 0 && !fpA)  
     wiener_coefficients(U,M,fRSigma2,p,npts);
     
     else if (strcmp(option,"ah") == 0 && !fpA)
     {
     
     float *est_pcs = new float[p];
     for (int ii=0; ii < p; ii++)		est_pcs[ii] = fpSPC[ (npts-1) * p + ii ] ;
     
     
     hard_principal_adaptive_values(S, M, est_pcs, fSigma, fRMult, p, npts);
     
     delete[] est_pcs;
     
     }
     else if (strcmp(option,"aw") == 0 && !fpA)
     {
     
     float *est_pcs = new float[p];
     for (int ii=0; ii < p; ii++)		est_pcs[ii] = fpSPC[ (npts-1) * p + ii ] ;
     
     wiener_principal_adaptive_values(S, M, est_pcs,  fSigma, fRMult, p, npts);
     
     delete[] est_pcs;
     
     }  else  if (strcmp(option,"w") == 0 && fpA)  
     wiener_oracle_principal_values(S,M,fRSigma2,p,npts);
     
     else  if (strcmp(option,"h") == 0 && fpA)  
     hard_principal_values(S,M,fRSigma2,p,npts);
     
     else  if (strcmp(option,"cw") == 0 && fpA)  
     wiener_oracle_coefficients(U02,M,fRSigma2,p,npts);
     
     else  if (strcmp(option,"ch") == 0 && fpA)  
     hard_coefficients(U02,M,fRSigma2,p,npts);
     
     
     
     
     for(int jj=0; jj < p; jj++)
     for(int ii=0; ii < npts; ii++)		
     U[ii][jj] *= M[ii][jj];
     
     
     
     
     
     /// Compute weight energy of coefficient modifiers
     if (ifAOverlap)
     {
     for(int ii=0; ii < npts; ii++)
     {
     float energy = 0.0; 
     for(int jj = 0; jj < p; jj++) energy += M[ii][jj] * M[ii][jj];
     
     if (energy > 1.0f)
     weight[ii] = 1.0 / energy;
     else 
     weight[ii] = 1.0;
     }
     }	
     
     
     
     /// Reconstruct denoised patches and save in XR2
     for(int ii=0; ii < npts; ii++)
     {
     for (int kk=0; kk < p; kk++)
     {
     XR2[ii][kk] = baricenter[kk] + scalar_product(U[ii], V[kk],p);
     }
     
     }
     
     
     
     } // END PCA Denoising for selected patches 
     
     
     
     
     
     /// Put back denoised values into image
     if ( ifCopying == 1 || ifCopying == 2 )
     {
     
     for(int ii=0; ii < npts; ii++)
     {
     int index = indexos[ii];
     int ci = index % iWidth;
     int cj = index / iWidth;
     
     
     // Put back denoised values
     int kk = 0;
     for (int ic = 0; ic < iChannels; ic++) 
     {
     for(int r=-iDWin; r <= iDWin; r++)
     for(int s=-iDWin; s <= iDWin; s++,kk++)
     {	
     int m = (cj + r) * iWidth + ci + s;    fpO[ic][m]   +=  weight[ii] * XR2[ii][kk];
     }
     }
     
     
     // Take into account number of values per pixel
     for(int r=-iDWin; r <= iDWin; r++)
     for(int s=-iDWin; s <= iDWin; s++,kk++)
     {	
     int m = (cj + r) * iWidth + ci + s;    mask[m] += weight[ii];
     }
     
     
     // Current pixel denoised as center of a block
     mask2[cj * iWidth + ci] += 1.0;
     
     }
     
     } else
     {
     
     // Put back denoised values
     int kk = 0;
     for (int ic = 0; ic < iChannels; ic++) 
     {
     for(int r=-iDWin; r <= iDWin; r++)
     for(int s=-iDWin; s <= iDWin; s++,kk++)
     {	
     int m = (y + r) * iWidth + x + s;    fpO[ic][m]   +=  weight[scurrent] * XR2[scurrent][kk];
     }
     }
     
     
     // Take into account number of values per pixel
     for(int r=-iDWin; r <= iDWin; r++)
     for(int s=-iDWin; s <= iDWin; s++,kk++)
     {	
     int m = (y + r) * iWidth + x + s;    mask[m] += weight[scurrent];
     }
     
     
     // Current pixel denoised as center of a block
     mask2[y * iWidth + x] += 1.0;
     
     }
     
     
     
     delete[] weight;
     
     
     }  // END for pixel x
     
     
     
     // delete allocated memory (in here because of parallel for y)
     delete[] baricenter;
     desallocate_float_matrix(X2, bsize, p);
     desallocate_float_matrix(XR2, bsize, p);	
     
     if (flag_computing_pca)
     {
     desallocate_float_matrix(M, bsize, p);
     desallocate_float_matrix(U, bsize, p);
     desallocate_float_matrix(V, p, p);
     }
     
     
     if (fpA && flag_computing_pca)
     {
     desallocate_float_matrix(U02, bsize, p);
     desallocate_float_matrix(X02, bsize, p);
     
     delete[] baricenter0;
     }
     
     
     delete[] tdistances;
     delete[] indexos;
     delete[] indexos2;
     
     
     
     }  // END for pixel y
     
     } // END of pragma
     
     /// Normalize values
     for(int i=0; i < iWidth*iHeight; i++)
     if (mask[i] > 0.0) {
     
     for (int ic=0; ic < iChannels; ic++) 
     fpO[ic][i] /= mask[i];
     
     }
     else {
     for (int ic=0; ic < iChannels; ic++) 
     fpO[ic][i] = fpI[ic][i];
     
     }	
     
     
     
     delete[] mask;
     delete[] mask2;
     
     }
     
     */
	
	
	
	
	
	float lmmse_set_threshold(int idisttype, int ifK, double fSigma2, double fMult, int p, double *tdistances, int ikt)
	{
		
		
		double iomeank, iostdk, difk;
		difk = compute_knn_mean_std(ifK, &iomeank, &iostdk, tdistances, ikt);
		double pthresh=0.0f;
		
		if (idisttype == 1)
			pthresh = fMult * fMult * (double) p;
		
		else if (idisttype == 11)
		{
			pthresh = fMult * fMult * (double) p;
			pthresh = MAX(pthresh, difk);
			
		}else if (idisttype == 2)
			pthresh = difk;
		
		else if (idisttype == 3)
			pthresh = MIN(fMult * fMult * (double) p , difk);
		
		else if (idisttype == 4)
		{
			float average = 2.0f * fSigma2 * (double) p;
			float variance = (8.0f * fSigma2 * fSigma2) / (double) (p);
			float std = (double) p * (double) sqrt(variance);
			
			pthresh = (double) average + fMult * std; 
		}		
		else if (idisttype == 5)
		{
			float average = 2.0f * fSigma2 * (double) p;
			float variance = (8.0f * fSigma2 * fSigma2) / (double) (p);
			float std = (double) p * (double) sqrt(variance);
			
			pthresh = MIN((double) average + fMult * std, difk);
			
		}		
		else if (idisttype == 6)
		{
			pthresh = iomeank + fMult * iostdk;
			
		}						
		else if (idisttype == 7)
		{
			float average = 2.0f * fSigma2 * (double) p;
			float variance = (8.0f * fSigma2 * fSigma2) / (double) (p);
			float std = (double) p * (double) sqrt(variance);
			
			pthresh = MAX(average + fMult * std, iomeank + fMult * iostdk );
			
		} else if (idisttype == 8)
		{
			
			float average = 2.0f * fSigma2 * (double) p;
			float variance = (8.0f * fSigma2 * fSigma2) / (double) (p);
			float std = (double) p * (double) sqrt(variance);
			
			pthresh = MAX(average + fMult * std, difk);
			
		}
		return pthresh;
	}
	
	
	
	
	float lmmse_compute_std(const laMatrix & X2, float *vmean, int npts, int p, int iChannels)
	{
		
		float std = 0.0f;
        float fValueAux = 0.0f;
        int iSize = p / iChannels;
        
		
		for (int cc = 0; cc < iChannels; cc++)
		{
			
			float mean=0.0f;
			float mean2=0.0f;
			
			for (int ii=0; ii < npts; ii++)
				for (int kk=0; kk < iSize ; kk++)
				{
                    fValueAux = X2[ii][cc * iSize + kk];
					mean += fValueAux;
					mean2 += fValueAux * fValueAux;
				}
			
			mean /= (float) (npts * iSize);
			mean2 /= (float) (npts * iSize);
			
			vmean[cc] = mean;
			
			std += sqrt(mean2 - mean * mean);
		}
		
		std /= (float) iChannels;
		return std;
	}
	
    
    
	
	void  lmmse_center_data(laMatrix  &X,double *baricenter)
	{
        
		int n = X.nrows();
        int p = X.ncols();
        
		
		for(int j=0;j<p;j++)
        {
			
			double some=0.0;
			for(int i=0; i<n; i++) some += X[i][j];
			
			some/=(double) n;
			baricenter[j]=some;
			
			for(int i=0; i<n; i++) X[i][j] -= baricenter[j];
		}
		
	}
	
	
    
    
    
	
	
    
    
    
    
    void lmmse_covariance_matrix(const laMatrix &X, laMatrix &Covariance)
	{
        int n = X.nrows();
        int p = X.ncols();
        
		
		for(int j=0;j<p;j++)
			for(int k=0;k<=j;k++)
            {
				
                float some=0.0;
				for(int i=0;i<n;i++)        some += X[i][j] * X[i][k];
				some/=(float) (n-1);
                
                Covariance[k][j]=some;
                Covariance[j][k]=some;
                
            }
		
	}
	
    
	
    
    //ifCopying:		0: Block denoised,  1: Entire Neighborhood denoised,  2: Entire neighborhood denoised and Copier-Coller	strategy 
	// iWin:			iWin x iWin window comparison
	// iBloc:			iBloc x iBloc research zone
	// fsigma:			noise std
	// idisttype		method for selecting similar blocks.  
	// iSampling
	// ifK				KNN if suitable idistype selected
	// ifUseFlat		Use flat
	// useFlatPar
	// ifSecondaryFilt		Secondary image used also for filtering
	void lmmse_filtering(float **fpI, float **fpA, float **fpO,
						 char *algoType,
						 int ifCopying,	
						 int iWinC, int iWin, int iBloc,
						 float fSigma, int idisttype,
						 int iSampling, int ifK, int ifuseFlat, float useFlatPar, int ifSecondaryFilt,
						 float fMult, float fRMult, int iChannels, int iWidth, int iHeight)
	{
		
		
		
		//! Parameters
        int n = (iWidth - iWin) * (iHeight - iWin);
		int p =  iWin * iWin * iChannels;	
        int pC = iWinC * iWinC * iChannels;	
        
		int iWin2 = iWin * iWin;
		
        int iDWin = (iWin-1) / 2;
        int iDWinC = (iWinC-1) / 2;
        
        int iDBloc = (iBloc-1) / 2;
		int bsize = iBloc * iBloc;
		
        
        printf("w: %d x %d \t bloc: %d x %d \t K: %d \t fRMult: %f \t useFlat: %d \t useSecondary: %d\n", iWin, iWin, iBloc, iBloc, ifK, fRMult, ifuseFlat, ifSecondaryFilt);
        
		//! Mask: weight sum per pixel
		float *mask = new float[iWidth*iHeight];
		fpClear(mask,0.0f,iWidth*iHeight);
		
		
		
		//! Mask2: treated pixels being central pixels of the window 
		//!        or copied by copy paste strategy being center of the window
		float *mask2 = new float[iWidth*iHeight];
		fpClear(mask2,0.0f,iWidth*iHeight);
		
		
		
		// clear output
		for (int ii=0; ii < iChannels; ii++) fpClear(fpO[ii], 0.0f, iWidth*iHeight);
		
		
#pragma omp parallel shared(fpI, fpO)
		{
			
			
#pragma omp for schedule(dynamic) nowait 
			
			
			for (int y = iDWinC; y < iHeight - iDWinC - 1; y+= iSampling)
			{
				
                
				//! Memory: data matrix with as maximum bsize samples
				laMatrix X2(bsize,p);				
				laVector baricenter(p);
				
				//!  Memory  with a secondary image
				laMatrix X02;
				laVector baricenter0(p);
				
				if (fpA && ifSecondaryFilt)
				{
					X02.create(bsize, p);
					//baricenter0.create(p);
				}
				
                
				
				//! Vector of squared distances 
				laVector tdistances(bsize);
				
				
				/// Other variables
				int scurrent = 0;							// position of reference pixel in the X2 table
				int *indexos = new int[n];					// indexos[k]: coordinates of block in k-entry of X2
				int *indexos2 = new int[n];					// indexos2[k]: position in neighborhood of k-entry of X2
				
				
				
				
				for (int x = iDWinC; x < iWidth - iDWinC - 1; x+=iSampling)
					if ( (ifCopying == 0) || (ifCopying == 1) || (ifCopying == 2 && mask2[y*iWidth+x] == 0.0f))		
					{																								
						
						//! Learning zone
						int imin=MAX(x-iDBloc,iDWinC);
						int jmin=MAX(y-iDBloc,iDWinC);
						int imax=MIN(x+iDBloc,iWidth-1-iDWinC);
						int jmax=MIN(y+iDBloc,iHeight-1-iDWinC);
						
						
						
						//! Computing distances
						tdistances = 0.0f;
						int ikt = 0;
						for(int j=jmin;j<=jmax;j++)
							for(int i=imin;i<=imax;i++,ikt++) 
							{
								
								float fDif;	
								if (fpA) 
									fDif  = fiL2FloatDist(fpA,fpA,x,y,i,j,iDWinC,iDWinC,iChannels, iWidth,iWidth);  
								
								else 
									fDif  = fiL2FloatDist(fpI,fpI,x,y,i,j,iDWinC,iDWinC,iChannels, iWidth,iWidth);  
								
								tdistances[ikt] = fDif;
							}
						
						
						
						//!  Set threshold for patches
						//! 1. Use fixed value
						//! 2. Use KNN
						//! 3. Use fixed value with at maximum k neighbours
						//! 4. Use average + m * std
						//! 5. Use average + m * std with at maximum k neighbours
						//! 6. Use average + m * std being average and std of the first KNN
						//! 7. Use the maximum of 4 and 5
						//! 8. Use the maximum of Average + fMult * std  and  KNN
						float pthresh=0.0;
						float fSigma2 = fSigma * fSigma;
						pthresh = lmmse_set_threshold(idisttype, ifK, fSigma2,fMult, pC, tdistances.v(), ikt);
						
						
                        
						/// Selecting similar patches
						int npts=0;
						ikt = 0;
						for(int j=jmin;j<=jmax;j++)
							for(int i=imin;i<=imax;i++,ikt++) 
							{
								
								float fDif = tdistances[ikt];
								if (fDif < pthresh)
								{
                                    
									// Fill X2 with noisy patch
									int kk = 0;
									for (int ii=0; ii < iChannels; ii++) 
									{
										for(int r=-iDWin; r <= iDWin; r++)
											for(int s=-iDWin; s <= iDWin; s++,kk++)
											{ 
												int m = (j + r) * iWidth + i + s;  X2[npts][kk] = fpI[ii][m];
											}	
									}
									
									
									// Fill X2 with additional patch if additional image selected
									if (fpA)
									{
										kk=0;
										for (int ii=0; ii < iChannels; ii++) 
										{
											for(int r=-iDWin; r <= iDWin; r++)
												for(int s=-iDWin; s <= iDWin; s++,kk++)   
												{ 
													int m = (j + r) * iWidth + i + s; X02[npts][kk] = fpA[ii][m]; 
												}
										}
									}	
									
									
									indexos[npts] = j*iWidth+i;					
									indexos2[npts] = ikt;	
									if (y == j && x == i) scurrent = npts;
									
									npts++;
								}
								
							}	
						
						
                        
						//! Resize X2
                        laMatrix X2res = X2.copyBlock(0, 0, npts, p);
                        
                        
						//! Resize secondary matrix
						laMatrix X02res;
						if (fpA &&  ifSecondaryFilt)
                            X02res = X02.copyBlock(0, 0, npts, p);
                        
                        
						//! Matrix of denoised vectors
						laMatrix XR2(npts, p); XR2 = 0.0;
                        
						
						//! Compute standard deviation for X2
						float *vmean = new float[iChannels];
						float lmmse_std = lmmse_compute_std(X2res, vmean, npts, p, iChannels);
						
						
						//! If use flat 
						if (ifuseFlat && lmmse_std < useFlatPar * fSigma )											
						{
							
							for(int ii=0; ii < npts; ii++)
								for (int cc=0; cc < iChannels ; cc++)
									for (int kk=0; kk < iWin2; kk++)
										XR2[ii][cc * iWin2 + kk] = vmean[cc];
							
						}  else if (strcmp(algoType,"bayes") == 0)  // Bayesian
						{
							
                            //! Compute covariance matrix
                            lmmse_center_data(X2res, baricenter.v());		
                            
                            laMatrix Covariance(p,p);
                            if (fpA &&  ifSecondaryFilt)
                            {
                                lmmse_center_data(X02res, baricenter0.v());
                                lmmse_covariance_matrix(X02res, Covariance);
                                
                            } else
                            {
                                lmmse_covariance_matrix(X2res, Covariance);
                            }
                            
                            
                            //! Modify diagonal of covariance matrix
                            float val_diag = fRMult * fSigma * fSigma;
                            laMatrix Covariance2 = Covariance;
                            if (fpA &&  ifSecondaryFilt)
                            {
                                for (int ii=0; ii < p; ii++) Covariance2[ii][ii] += val_diag;
                                
                            } else
                            {
                                for (int ii=0; ii < p; ii++) Covariance2[ii][ii] -= val_diag;
                            }
                            
                            
							
							
							laMatrix X2tres = X2res.transposed();
							laMatrix inverse(p,p);
							if (fpA && ifSecondaryFilt)
							{
								
								luinv(Covariance2, inverse);
								X2tres = inverse * X2tres;
								X2tres = Covariance * X2tres;
								
								
							} else {
								
								luinv(Covariance, inverse);
								X2tres = inverse * X2tres;
								X2tres = Covariance2 * X2tres;
							}
                            
                            
							
                            
                            
							
                            
                            //! Add baricenter
                            for (int jj = 0; jj < npts; jj++)
                            {
                                for (int ii = 0; ii < p; ii++)
                                    XR2[jj][ii] =  baricenter[ii] + X2tres[ii][jj];
                            }
                            
                            
                            // END Bayes Denoising for selected patches 	
						} 
						else if ( strcmp(algoType,"pcaThr") == 0 || strcmp(algoType,"pcaWie") == 0 || strcmp(algoType,"pcaCThr") == 0 || strcmp(algoType,"pcaCWie") == 0)  // NLPCA
						{
                            
							
							
							//! Compute covariance matrix
							lmmse_center_data(X2res, baricenter.v());		
							
							laVector S(p);
							laMatrix V(p, p);
							laMatrix U(npts, p);
							laMatrix M(npts, p);
							laMatrix U02;
							
							
							if (fpA &&  ifSecondaryFilt)
							{
                                
								U02.create(npts, p);
								
								lmmse_center_data(X02res, baricenter0.v());
								compute_pca_svd(X02res, S, V, U02);
								
								
								// Compute U coefficients with PCA basis V
								for(int ii=0; ii < npts; ii++)
								{ 
									for(int jj=0; jj < p; jj++)
									{
										float aux = 0.0; 
										for (int kk=0; kk <p; kk++) 
											aux += X2res[ii][kk] * V[kk][jj]; U[ii][jj] =  aux; 
									}
								}
								
								
							} else
							{
								compute_pca_svd(X2res, S, V, U);
							}
							
							
							
							//! Modify Coefficients
							
                            float fRSigma2 = fRMult * fRMult * fSigma2;
							
							
							if (strcmp(algoType,"pcaWie") == 0 && !fpA)  
								wiener_principal_values(S.v(),M.v(),fRSigma2,p,npts);
							
							else if (strcmp(algoType,"pcaThr") == 0 && !fpA) 
								hard_principal_values(S.v(),M.v(),fRSigma2,p,npts);
							
							else  if (strcmp(algoType,"pcaCWie") == 0 && fpA)  
								wiener_oracle_coefficients(U02.v(),M.v(),fRSigma2,p,npts);
							
							else  if (strcmp(algoType,"pcaCThr") == 0 && fpA)  
								hard_coefficients(U02.v(),M.v(),fRSigma2,p,npts);
							
							
							
							for(int jj=0; jj < p; jj++)
								for(int ii=0; ii < npts; ii++)		
									U[ii][jj] *= M[ii][jj];
							
							
							
							/// Reconstruct denoised patches and save in XR2
							for(int ii=0; ii < npts; ii++)
							{
								for (int kk=0; kk < p; kk++)
								{
									XR2[ii][kk] = baricenter[kk] + scalar_product(U[ii], V[kk],p);
								}
								
							}
							
							
							
						} else if (strcmp(algoType,"bayes2") == 0)
						{
                            
							
							
							//! Compute covariance matrix
							lmmse_center_data(X2res, baricenter.v());		
							
							laVector S(p);
							laMatrix V(p, p);
							laMatrix U(npts, p);
							laMatrix M(npts, p);
							laMatrix U02;							
							
							
							/*if (fpA &&  ifSecondaryFilt)
							{
                                
								U02.create(npts, p);
								
								lmmse_center_data(X02res, baricenter0.v());
								compute_pca_svd(X02res.v(), S.v(), V.v(), U02.v(), npts, p);
								
								
								// Compute U coefficients with PCA basis V
								for(int ii=0; ii < npts; ii++)
								{ 
									for(int jj=0; jj < p; jj++)
									{
										float aux = 0.0; 
										for (int kk=0; kk <p; kk++) 
											aux += X2res[ii][kk] * V[kk][jj]; U[ii][jj] =  aux; 
									}
								}
								
								
							} else
							{*/
								compute_pca_svd(X2res, S, V, U);
							//}
							
							
							
							//! Modify Coefficients
                            float fRSigma2 = fRMult * fRMult * fSigma2;
														
							
							for(int jj=0; jj < p; jj++)
								for(int ii=0; ii < npts; ii++)		
										if (S[jj] > 0.001)  M[ii][jj] = fRSigma2 / S[jj];
										else M[ii][jj] = 0.0f;
									
							
							for(int jj=0; jj < p; jj++)
								for(int ii=0; ii < npts; ii++)		
									U[ii][jj] *= M[ii][jj];
							
							
							
							/// Reconstruct denoised patches and save in XR2
							for(int ii=0; ii < npts; ii++)
							{
								for (int kk=0; kk < p; kk++)
								{
									XR2[ii][kk] = baricenter[kk]  + X2res[ii][kk]  - scalar_product(U[ii], V[kk],p);
								}
								
							}
							
							
							
						} else
						{
							printf("error :: lmmse denosing :: select a correct type algorithm\n");
						}
						
                        
                        
						
						
						/// Put back denoised values into image
						if ( ifCopying == 1 || ifCopying == 2 )
						{
							
							for(int ii=0; ii < npts; ii++)
							{
								int index = indexos[ii];
								int ci = index % iWidth;
								int cj = index / iWidth;
								
								
								// Put back denoised values
								int kk = 0;
								for (int ic = 0; ic < iChannels; ic++) 
								{
									for(int r=-iDWin; r <= iDWin; r++)
										for(int s=-iDWin; s <= iDWin; s++,kk++)
										{	
											int m = (cj + r) * iWidth + ci + s;    fpO[ic][m]   +=    XR2[ii][kk];
										}
								}
								
								
								// Take into account number of values per pixel
								for(int r=-iDWin; r <= iDWin; r++)
									for(int s=-iDWin; s <= iDWin; s++,kk++)
									{	
										int m = (cj + r) * iWidth + ci + s;    mask[m]++;
									}
								
								
								// Current pixel denoised as center of a block
								mask2[cj * iWidth + ci] += 1.0;
								
							}
							
						} else
						{
							
							// Put back denoised values
							int kk = 0;
							for (int ic = 0; ic < iChannels; ic++) 
							{
								for(int r=-iDWin; r <= iDWin; r++)
									for(int s=-iDWin; s <= iDWin; s++,kk++)
									{	
										int m = (y + r) * iWidth + x + s;    fpO[ic][m]   +=   XR2[scurrent][kk];
									}
							}
							
							
							// Take into account number of values per pixel
							for(int r=-iDWin; r <= iDWin; r++)
								for(int s=-iDWin; s <= iDWin; s++,kk++)
								{	
									int m = (y + r) * iWidth + x + s;    mask[m]++;
								}
							
							
							// Current pixel denoised as center of a block
							mask2[y * iWidth + x] += 1.0;
							
						}
						
                        
					}  // END for pixel x
				
				
				
                
				delete[] indexos;
				delete[] indexos2;
				
				
				
			}  // END for pixel y
			
		} // END of pragma
		
		
		/// Normalize values
		for(int i=0; i < iWidth*iHeight; i++)
			if (mask[i] > 0.0) {
				
				for (int ic=0; ic < iChannels; ic++) 
					fpO[ic][i] /= mask[i];
				
			}
			else {
				for (int ic=0; ic < iChannels; ic++) 
					fpO[ic][i] = fpI[ic][i];
				
			}	
		
		
		
		delete[] mask;
		delete[] mask2;
		
	}
	
	
    
    
    
    
    
    
    
    //ifCopying:		0: Block denoised,  1: Entire Neighborhood denoised,  2: Entire neighborhood denoised and Copier-Coller					strategy 
	// iWin:			iWin x iWin window comparison
	// iBloc:			iBloc x iBloc research zone
	// fsigma:			noise std
	// idisttype		method for selecting similar blocks.  
	// iSampling
	// ifK				KNN if suitable idistype selected
	// ifUseFlat		Use flat
	// useFlatPar
	// ifSecondaryFilt		Secondary image used also for filtering
	void lmmse_filtering_weighted_window(float **fpI, float **fpA, float **fpO,
                                         char *algoType,
                                         int ifCopying,	
                                         int iWin, int jWin, float *kernel,
                                         int iBloc,
                                         float fSigma, int idisttype,
                                         int iSampling, int ifK, int ifuseFlat, float useFlatPar, int ifSecondaryFilt,
                                         float fMult, float fRMult, int iChannels, int iWidth, int iHeight)
	{
		
        
		
		//! Parameters
        int n = (iWidth - iWin) * (iHeight - jWin);
        int ijWin = iWin * jWin;
        
        int iDWin = (iWin-1) / 2;
        int jDWin = (jWin-1) / 2;
        
        
        
        
        int iDBloc = (iBloc-1) / 2;
		int bsize = iBloc * iBloc;
		
        // Adapted Value of p
        // int p = iWin * iWin * iChannels;	
		int p=0;
        for (int ii=0; ii < iWin * jWin ; ii++)  if (kernel[ii] > 0.0f) p++;
        p *= iChannels;
        
        
        printf("w: %d x %d \t p: %d \t bloc: %d x %d \t K: %d \t fRMult: %f \t useFlat: %d \t useSecondary: %d\n", iWin, jWin, p, iBloc, iBloc, ifK, fRMult, ifuseFlat, ifSecondaryFilt);
        
        
        
		//! Mask: weight sum per pixel
		float *mask = new float[iWidth*iHeight];
		fpClear(mask,0.0f,iWidth*iHeight);
		
		
		
		//! Mask2: treated pixels being central pixels of the window 
		//!        or copied by copy paste strategy being center of the window
		float *mask2 = new float[iWidth*iHeight];
		fpClear(mask2,0.0f,iWidth*iHeight);
		
		
		
		// clear output
		for (int ii=0; ii < iChannels; ii++) fpClear(fpO[ii], 0.0f, iWidth*iHeight);
		
		
#pragma omp parallel shared(fpI, fpO)
		{
			
			
#pragma omp for schedule(dynamic) nowait 
			
			
			for (int y = jDWin; y < iHeight - jDWin - 1; y+= iSampling)
			{
				
                
				//! Memory: data matrix with as maximum bsize samples
				laMatrix X2(bsize,p);				
				laVector baricenter(p);
				
				//!  Memory  with a secondary image
				laMatrix X02;
				laVector baricenter0(p);
				
				if (fpA && ifSecondaryFilt)
				{
					X02.create(bsize, p);
					//baricenter0.create(p);
				}
				
                
				
				//! Vector of squared distances 
				laVector tdistances(bsize);
				
				
				/// Other variables
				int scurrent = 0;							// position of reference pixel in the X2 table
				int *indexos = new int[n];					// indexos[k]: coordinates of block in k-entry of X2
				int *indexos2 = new int[n];					// indexos2[k]: position in neighborhood of k-entry of X2
				
				
				
				
				for (int x = iDWin; x < iWidth - iDWin - 1; x+=iSampling)
					if ( (ifCopying == 0) || (ifCopying == 1) || (ifCopying == 2 && mask2[y*iWidth+x] == 0.0f))		
					{																								
						
						//! Learning zone
						int imin=MAX(x-iDBloc,iDWin);
						int jmin=MAX(y-iDBloc,jDWin);
						int imax=MIN(x+iDBloc,iWidth-1-iDWin);
						int jmax=MIN(y+iDBloc,iHeight-1-jDWin);
						
						
						
						//! Computing distances
						tdistances = 0.0f;
						int ikt = 0;
						for(int j=jmin;j<=jmax;j++)
							for(int i=imin;i<=imax;i++,ikt++) 
							{
								
								float fDif;	
								if (fpA) 
									fDif  = fiL2FloatWDist(fpA,fpA,x,y,i,j,iDWin, jDWin, kernel, iChannels, iWidth,iWidth);  
								else 
									fDif  = fiL2FloatWDist(fpI,fpI,x,y,i,j,iDWin, jDWin, kernel,  iChannels, iWidth,iWidth);  
								
								tdistances[ikt] = fDif;
							}
						
						
						
						//!  Set threshold for patches
						//! 1. Use fixed value
						//! 2. Use KNN
						//! 3. Use fixed value with at maximum k neighbours
						//! 4. Use average + m * std
						//! 5. Use average + m * std with at maximum k neighbours
						//! 6. Use average + m * std being average and std of the first KNN
						//! 7. Use the maximum of 4 and 5
						//! 8. Use the maximum of Average + fMult * std  and  KNN
						float pthresh=0.0;
						float fSigma2 = fSigma * fSigma;
						pthresh = lmmse_set_threshold(idisttype, ifK, fSigma2,fMult, p, tdistances.v(), ikt);
						
						
                        
						/// Selecting similar patches
						int npts=0;
						ikt = 0;
						for(int j=jmin;j<=jmax;j++)
							for(int i=imin;i<=imax;i++,ikt++) 
							{
								
								float fDif = tdistances[ikt];
								if (fDif < pthresh)
								{
                                    
									// Fill X2 with noisy patch
									int kk = 0;
									for (int ii=0; ii < iChannels; ii++) 
									{
										for(int r=-jDWin; r <= jDWin; r++)
											for(int s=-iDWin; s <= iDWin; s++)	
                                                if (kernel[ (jDWin + r) * iWin + s + iDWin] > 0.0f)
                                                { 
                                                    int m = (j + r) * iWidth + i + s;  X2[npts][kk] = fpI[ii][m]; kk++;
                                                }	
									}
									
                                    //printf("%d: (%d,%d) \t length = %d \n",npts, i, j, kk);
									//for (int iii = 0; iii < p; iii++) printf("%f \t", X2[npts][iii]);
                                    //printf("\n");
                                    
									// Fill X2 with additional patch if additional image selected
									if (fpA)
									{
										kk=0;
										for (int ii=0; ii < iChannels; ii++) 
										{
											for(int r=-jDWin; r <= jDWin; r++)
												for(int s=-iDWin; s <= iDWin; s++)   
                                                    if (kernel[ (jDWin + r) * iWin + s + iDWin] > 0.0f)
                                                    { 
                                                        int m = (j + r) * iWidth + i + s; X02[npts][kk] = fpA[ii][m]; kk++;
                                                    }
										}
									}	
									
									
									indexos[npts] = j*iWidth+i;					
									indexos2[npts] = ikt;	
									if (y == j && x == i) scurrent = npts;
									
									npts++;
								}
								
							}	
						
						
                        
						//! Resize X2
                        laMatrix X2res = X2.copyBlock(0, 0, npts, p);
                        
                        
						//! Resize secondary matrix
						laMatrix X02res;
						if (fpA &&  ifSecondaryFilt)
                            X02res = X02.copyBlock(0, 0, npts, p);
                        
                        
						//! Matrix of denoised vectors
						laMatrix XR2(npts, p); XR2 = 0.0;
                        
						
						//! Compute standard deviation for X2
						float *vmean = new float[iChannels];
						float lmmse_std = lmmse_compute_std(X2res, vmean, npts, p, iChannels);
						
						
						//! If use flat 
						if (ifuseFlat && lmmse_std < useFlatPar * fSigma )											
						{
							int iSize = p / iChannels;
                            
							for(int ii=0; ii < npts; ii++)
								for (int cc=0; cc < iChannels ; cc++)
									for (int kk=0; kk < iSize; kk++)
										XR2[ii][cc * iSize + kk] = vmean[cc];
							
						}  else if (strcmp(algoType,"bayes") == 0)  // Bayesian
						{
							
                            //! Compute covariance matrix
                            lmmse_center_data(X2res, baricenter.v());		
                            
                            laMatrix Covariance(p,p);
                            if (fpA &&  ifSecondaryFilt)
                            {
                                lmmse_center_data(X02res, baricenter0.v());
                                lmmse_covariance_matrix(X02res, Covariance);
                                
                            } else
                            {
                                lmmse_covariance_matrix(X2res, Covariance);
                            }
                            
                            
                            //! Modify diagonal of covariance matrix
                            float val_diag = fRMult * fSigma * fSigma;
                            laMatrix Covariance2 = Covariance;
                            if (fpA &&  ifSecondaryFilt)
                            {
                                for (int ii=0; ii < p; ii++) Covariance2[ii][ii] += val_diag;
                                
                            } else
                            {
                                for (int ii=0; ii < p; ii++) Covariance2[ii][ii] -= val_diag;
                            }
                            
                            
							
							
							laMatrix X2tres = X2res.transposed();
							laMatrix inverse(p,p);
							if (fpA && ifSecondaryFilt)
							{
								
								luinv(Covariance2, inverse);
								X2tres = inverse * X2tres;
								X2tres = Covariance * X2tres;
								
								
							} else {
								
								luinv(Covariance, inverse);
								X2tres = inverse * X2tres;
								X2tres = Covariance2 * X2tres;
							}
                            
                            
							
                            
                            
							
                            
                            //! Add baricenter
                            for (int jj = 0; jj < npts; jj++)
                            {
                                for (int ii = 0; ii < p; ii++)
                                    XR2[jj][ii] =  baricenter[ii] + X2tres[ii][jj];
                            }
                            
                            
                            // END Bayes Denoising for selected patches 	
						} 
						else if ( strcmp(algoType,"pcaThr") == 0 || strcmp(algoType,"pcaWie") == 0 || strcmp(algoType,"pcaCThr") == 0 || strcmp(algoType,"pcaCWie") == 0)  // NLPCA
						{
                            
							
							
							//! Compute covariance matrix
							lmmse_center_data(X2res, baricenter.v());		
							
							laVector S(p);
							laMatrix V(p, p);
							laMatrix U(npts, p);
							laMatrix M(npts, p);
							laMatrix U02;
							
							
							if (fpA &&  ifSecondaryFilt)
							{
                                
								U02.create(npts, p);
								
								lmmse_center_data(X02res, baricenter0.v());
								compute_pca_svd(X02res, S, V, U02);
								
								
								// Compute U coefficients with PCA basis V
								for(int ii=0; ii < npts; ii++)
								{ 
									for(int jj=0; jj < p; jj++)
									{
										float aux = 0.0; 
										for (int kk=0; kk <p; kk++) 
											aux += X2res[ii][kk] * V[kk][jj]; U[ii][jj] =  aux; 
									}
								}
								
								
							} else
							{
								compute_pca_svd(X2res, S, V, U);
							}
							
							
							
							//! Modify Coefficients
							
                            float fRSigma2 = fRMult * fRMult * fSigma2;
							
							
							if (strcmp(algoType,"pcaWie") == 0 && !fpA)  
								wiener_principal_values(S.v(),M.v(),fRSigma2,p,npts);
							
							else if (strcmp(algoType,"pcaThr") == 0 && !fpA) 
								hard_principal_values(S.v(),M.v(),fRSigma2,p,npts);
							
							else  if (strcmp(algoType,"pcaCWie") == 0 && fpA)  
								wiener_oracle_coefficients(U02.v(),M.v(),fRSigma2,p,npts);
							
							else  if (strcmp(algoType,"pcaCThr") == 0 && fpA)  
								hard_coefficients(U02.v(),M.v(),fRSigma2,p,npts);
							
							
							
							for(int jj=0; jj < p; jj++)
								for(int ii=0; ii < npts; ii++)		
									U[ii][jj] *= M[ii][jj];
							
							
							
							/// Reconstruct denoised patches and save in XR2
							for(int ii=0; ii < npts; ii++)
							{
								for (int kk=0; kk < p; kk++)
								{
									XR2[ii][kk] = baricenter[kk] + scalar_product(U[ii], V[kk],p);
								}
								
							}
							
							
							
						} else
						{
							printf("error :: lmmse denosing :: select a correct type algorithm\n");
						}
						
                        
                        
						
						
						/// Put back denoised values into image
						if ( ifCopying == 1 || ifCopying == 2 )
						{
							
							for(int ii=0; ii < npts; ii++)
							{
								int index = indexos[ii];
								int ci = index % iWidth;
								int cj = index / iWidth;
								
								
								// Put back denoised values
								int kk = 0;
								for (int ic = 0; ic < iChannels; ic++) 
								{
									for(int r=-jDWin; r <= jDWin; r++)
										for(int s=-iDWin; s <= iDWin; s++)
                                            if (kernel[ (jDWin + r) * iWin + s + iDWin] > 0.0f)
                                            {	
                                                int m = (cj + r) * iWidth + ci + s;    fpO[ic][m]   +=    XR2[ii][kk]; kk++;
                                            }
								}
								
								
								// Take into account number of values per pixel
								for(int r=-jDWin; r <= jDWin; r++)
									for(int s=-iDWin; s <= iDWin; s++)
                                        if (kernel[ (jDWin + r) * iWin + s + iDWin] > 0.0f)
                                        {	
                                            int m = (cj + r) * iWidth + ci + s;    mask[m]++;
                                        }
								
								
								// Current pixel denoised as center of a block
								mask2[cj * iWidth + ci] += 1.0;
								
							}
							
						} else
						{
							
							// Put back denoised values
							int kk = 0;
							for (int ic = 0; ic < iChannels; ic++) 
							{
								for(int r=-jDWin; r <= jDWin; r++)
									for(int s=-iDWin; s <= iDWin; s++)
                                        if (kernel[ (jDWin + r) * iWin + s + iDWin] > 0.0f)
                                        {	
                                            int m = (y + r) * iWidth + x + s;    fpO[ic][m]   +=   XR2[scurrent][kk]; kk++;
                                        }
							}
							
							
							// Take into account number of values per pixel
							for(int r=-jDWin; r <= jDWin; r++)
								for(int s=-iDWin; s <= iDWin; s++)
                                    if (kernel[ (jDWin + r) * iWin + s + iDWin] > 0.0f)
                                    {	
                                        int m = (y + r) * iWidth + x + s;    mask[m]++;
                                    }
							
							
							// Current pixel denoised as center of a block
							mask2[y * iWidth + x] += 1.0;
							
						}
						
                        
					}  // END for pixel x
				
				
				
                
				delete[] indexos;
				delete[] indexos2;
				
				
				
			}  // END for pixel y
			
		} // END of pragma
		
		
		/// Normalize values
		for(int i=0; i < iWidth*iHeight; i++)
			if (mask[i] > 0.0) {
				
				for (int ic=0; ic < iChannels; ic++) 
					fpO[ic][i] /= mask[i];
				
			}
			else {
				for (int ic=0; ic < iChannels; ic++) 
					fpO[ic][i] = fpI[ic][i];
				
			}	
		
		
		
		delete[] mask;
		delete[] mask2;
		
	}
	
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    void nlmeans_single_frame(int iIndex,	// iIndex of denoised frame
                              int iDtemp,	// Half size of temporal window
                              int iDWin,		// Half size of comparison window  
                              int iDBloc,		// Half size of research window
                              float fSigma,		// Noise parameter
                              float fFiltPar,	// Filtering parameter
                              float ***fpI,		// Input
                              float **fpO,		// Output
                              float *fpMask,	// Mask
                              int iFrames, int iChannels, int iWidth,int iHeight)
    {
        
        
        // length of each channel
        int iwxh = iWidth * iHeight;
        
        
        //  length of comparison window
        int ihwl = (2*iDWin+1);	
        int iwl = (2*iDWin+1) * (2*iDWin+1);
        int icwl = iChannels * iwl;
        
        
        
        // filtering parameter
        // float fSigma2 = fSigma * fSigma;	
        float fH = fFiltPar * fSigma;
        float fH2 = fH * fH;
        
        // multiply by size of comparison window, since distances are not normalized
        fH2 *= (float) icwl;
        
        
        
        // tabulate exp(-x), faster than using directly function expf
        int iLutLength = (int) rintf((float) LUTMAX * (float) LUTPRECISION);
        float *fpLut = new float[iLutLength];
        wxFillExpLut(fpLut,iLutLength);
        
        
        
        
        
        // PROCESS STARTS
        // for each frame
        
        
        // auxiliary variable
        // number of estimated values per pixel
        float *fpCount = new float[iwxh];
        fpClear(fpCount, 0.0f,iwxh);
        
        
        
        
        
        // PROCESS STARTS
        // for each pixel (x,y)
#pragma omp parallel shared(fpI, fpO)
        {
            
            
#pragma omp for schedule(dynamic) nowait 
            
            
            for(int y=0; y < iHeight ; y++)
            {
                
                // auxiliary variable
                // denoised patch centered at a certain pixel
                float **fpODenoised = new float*[iChannels];
                for (int ii=0; ii < iChannels; ii++) fpODenoised[ii] = new float[iwl];
                
                
                for(int x=0 ; x < iWidth;  x++)
                    if (fpMask[y * iWidth + x] > 0.0f)
                    {
                        
                        // reduce the size of the comparison window if we are near the boundary  
                        int iDWin0 = MIN(iDWin,MIN(iWidth-1-x,MIN(iHeight-1-y,MIN(x,y))));
                        
                        
                        // research zone depending on the boundary and the size of the window
                        int tmin=MAX(iIndex-iDtemp,0);
                        int tmax=MIN(iIndex+iDtemp,iFrames-1);
                        
                        int imin=MAX(x-iDBloc,iDWin0);
                        int jmin=MAX(y-iDBloc,iDWin0);
                        
                        int imax=MIN(x+iDBloc,iWidth-1-iDWin0);
                        int jmax=MIN(y+iDBloc,iHeight-1-iDWin0);
                        
                        
                        
                        //  clear current denoised patch
                        for (int ii=0; ii < iChannels; ii++) fpClear(fpODenoised[ii], 0.0f, iwl);
                        
                        
                        
                        // maximum of weights for central pixel in case exponential strategy is used
                        float fMaxWeight = 0.0f;
                        
                        
                        // sum of weights
                        float fTotalWeight = 0.0f;
                        
                        
                        
                        for(int t=tmin; t <= tmax; t++)
                            for(int j=jmin; j <= jmax; j++)
                                for(int i=imin ; i <= imax; i++) 
                                    if (i!=x || j!=y)
                                    {
                                        
                                        float fDif = fiL2FloatDist(fpI[iIndex],fpI[t],x,y,i,j,iDWin0,iDWin0,iChannels,iWidth,iWidth);
                                        fDif = MAX(fDif /*- 2.0f * (float) icwl *  fSigma2*/, 0.0f);
                                        fDif = fDif / fH2;	
                                        
                                        float fWeight = wxSLUT(fDif,fpLut);
                                        
                                        if (fWeight > fMaxWeight) fMaxWeight = fWeight;
                                        
                                        fTotalWeight += fWeight;	
                                        
                                        
                                        
                                        float **cfpI = fpI[t];
                                        for(int is=-iDWin0; is <=iDWin0; is++)
                                        {	
                                            int aiindex = (iDWin+is) * ihwl + iDWin;
                                            int ail = (j+is)*iWidth+i;
                                            
                                            for(int ir=-iDWin0; ir <= iDWin0; ir++)
                                            {
                                                
                                                int iindex = aiindex + ir;
                                                int il= ail +ir;
                                                
                                                for(int ii=0; ii < iChannels; ii++)
                                                    fpODenoised[ii][iindex] += fWeight * cfpI[ii][il];
                                                
                                            }
                                            
                                        }
                                        
                                        
                                        
                                        
                                    }
                        
                        
                        // current point (x,y) with fMaxWeight
                        float **cfpI = fpI[iIndex];
                        for(int is=-iDWin0; is <=iDWin0; is++)
                        {
                            int aiindex = (iDWin+is) * ihwl + iDWin;
                            int ail=(y+is)*iWidth+x;
                            
                            for(int ir=-iDWin0; ir <= iDWin0; ir++)
                            {
                                
                                int iindex = aiindex + ir;
                                int il=ail+ir;
                                
                                for(int ii=0; ii < iChannels; ii++)
                                    fpODenoised[ii][iindex] += fMaxWeight * cfpI[ii][il];
                                
                            }
                        }
                        
                        
                        
                        fTotalWeight += fMaxWeight;	
                        
                        
                        // normalize average value
                        // else case treated as fpCount[y*width+x] = 0
                        float **cfpO = fpO; 
                        if (fTotalWeight > fTiny) 
                        {
                            
                            
                            
                            for(int is=-iDWin0; is <=iDWin0; is++)
                            {
                                int aiindex = (iDWin+is) * ihwl + iDWin;
                                int ail=(y+is)*iWidth+x;
                                
                                for(int ir=-iDWin0; ir <= iDWin0; ir++)
                                {
                                    int iindex = aiindex + ir;
                                    int il=ail+ ir;
                                    
                                    fpCount[il]++;
                                    
                                    for(int ii=0; ii < iChannels; ii++)
                                    {
                                        cfpO[ii][il] += fpODenoised[ii][iindex] / fTotalWeight; 
                                        
                                    }
                                    
                                }
                            }
                            
                            
                        }
                        
                        
                        
                        
                    }
                
                
            }
            
        }
        
        
        float **cfpO = fpO;
        float **cfpI = fpI[iIndex];
        
        for(int ii=0; ii < iwxh;ii++)
            if (fpCount[ii]>0.0)
            {
                for (int jj=0; jj < iChannels; jj++)  cfpO[jj][ii] /= fpCount[ii];
                
            }	else 
            {
                
                for (int jj=0; jj < iChannels; jj++)  cfpO[jj][ii] = cfpI[jj][ii];
            }
        
        
        
        
        
        
        
        
    }
    
    
    
    
}

